// BSP : STM32 boards groups by STMicroelectronics 2.7.1
// Borad select : STM Boards groups / LoRa boars
// Board part number : LoRa-E5 mini (Board to be selected from Tools submenu 'Board part number)

// 1. transmit data 12 bytes(Slave UID, mills(), Vbatt)
// 2. wait for response from Master
// 3. receive  data 12 bytes(Master UID, rssi, snr, reserved)
// 4. shutdown and wakeup

// 2024/02/11

#include <Arduino.h>
#include <Adafruit_GFX.h>       // 1.11.9
#include <Adafruit_SSD1306.h>   // 2.5.9
#include "myFonts/tahoma5pt7b.h"  // Any font
#include "myFonts/tahoma7pt7b.h"
#include <RadioLib.h>           // 6.4.2  https://github.com/jgromes/RadioLib
#include <STM32LowPower.h>      // 1.2.5  https://github.com/stm32duino/STM32LowPower
#include <STM32RTC.h>           // 1.4.0  https://github.com/stm32duino/STM32RTC

// BAT_pad --> 1000k --> A4_pin --> 510k --> D0_pin
#define PIN_VBATT      A4       // battery voltage measurement pin
#define READ_ENABLE    D0       // read PIN_VBATT enable LOW:enable, HIGH:disable

// RM0461 Reference manual Rev 8  37.1.4 IEEE 64-bit unique device ID register (UID64)
#define SLAVE_UID   *((uint32_t*)0x1FFF7580)    // use uid64 dor Slave device UID
#define MASTER_UID  0x0535EE2F  // sample Master device UID

Adafruit_SSD1306 display(128, 64, &Wire, -1);

// \Arduino15\packages\STMicroelectronics\hardware\stm32\2.7.1
// \variants\STM32WLxx\WL54JCI_WL55JCI_WLE4J(8-B-C)I_WLE5J(8-B-C)I\variant_LORA_E5_MINI.h : 173
STM32WLx radio = new STM32WLx_Module();
static const uint32_t rfswitch_pins[] =
                         {PA4,  PA5,  RADIOLIB_NC};
static const Module::RfSwitchMode_t rfswitch_table[] = {
  {STM32WLx::MODE_IDLE,  {LOW,  LOW}},
  {STM32WLx::MODE_RX,    {HIGH, LOW}},
//  {STM32WLx::MODE_TX_HP, {LOW, HIGH}},  // for LoRa-E5 mini
  {STM32WLx::MODE_TX_LP, {HIGH, HIGH}},   // for LoRa-E5-LE mini  
  END_OF_MODE_TABLE,
};

#define dataNum   12                // data bytes
union unionData {                   // for data type conversion
  uint32_t  dataBuff32[dataNum/4];
  float     dataBuff32F[dataNum/4];
  int16_t   dataBuff16[dataNum/2];
  uint8_t   dataBuff8[dataNum];
};
union unionData ud;

int rssi;               // signal RSSI [dBm]
int snr;                // signal SN ratio [dB]
int bps;                // data rate [bps]
uint32_t slaveuid = SLAVE_UID;    // Slave device UID
uint32_t masteruid;     // Master device UID
uint16_t sec;           // mS/1000 data [sec]
float Vbatt;            // battery voltage data [V]
char printBuff[4];      // for sprintf()
String  txdata = "";    // transmission data packet string
String  rxdata = "";    // received data packet string
bool rxstatus = false;  // receive status flag

bool receivedFlag = false;     // flag that a packet was received
bool transmittedFlag = false;  // flag that a packet was transmitted

// *******************************************************************************************************
void setup() {
  Serial.begin(115200);
  while(!Serial);
  delay(1000);
  Serial.println("**** Wake up ****");

  // Pin initialization
  pinMode(LED_RED, OUTPUT);         // built-in LED
  pinMode(READ_ENABLE, OUTPUT);     // read PIN_VBATT enable
  digitalWrite(LED_RED, HIGH);      // LED off
  digitalWrite(READ_ENABLE, LOW);   // read enable

  // ADC initialization
  analogReadResolution(12);         // ADC 12 bits

  // Display initialization
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.clearDisplay();
  display.setFont(&tahoma7pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("LoRa-E5");
  display.setCursor(0, 31);
  display.println("P2P Test");
  display.setCursor(0, 47);
  display.println("Slave");  
  display.display();
  delay(2000);
  display.clearDisplay();  


  // Radio initialization
  radio.setRfSwitchTable(rfswitch_pins, rfswitch_table);

  // int16_t begin(float freq = 434.0, float bw = 125.0, uint8_t sf = 9, uint8_t cr = 7, \
  // uint8_t syncWord = RADIOLIB_SX126X_SYNC_WORD_PRIVATE, int8_t power = 10, \
  // uint16_t preambleLength = 8, float tcxoVoltage = 1.6, bool useRegulatorLDO = false);
  int state = radio.begin(922.6, 125.0, 12, 5, 0x12, 13, 8, 1.7, 0);

  if (state == RADIOLIB_ERR_NONE) {
    Serial.println("radio.begin() success!");
  }
  else {
    Serial.print("failed, code ");
    Serial.println(state);
    while(true) {
      digitalWrite(LED_RED, LOW);
      delay(100);
      digitalWrite(LED_RED, HIGH);
      delay(100);      
    }
  }

  // callback function when received or transmitted a packet
  radio.setDio1Action(setFlag);

  // setup end 
  display.setCursor(0, 47);
  display.print("LoRa-E5 Slave");
  display.setCursor(0, 63);
  display.print("Setup END");
  display.display();
  delay(2000);
  display.clearDisplay();

  // Configure low power
  LowPower.begin();
}


// this function is called when a complete packet is received or transmitted
void setFlag(void) 
{
  uint16_t irqstatus = radio.getIrqStatus();
  if(irqstatus == RADIOLIB_SX126X_IRQ_RX_DONE) {
    receivedFlag = true;
  }
  else if(irqstatus == RADIOLIB_SX126X_IRQ_TX_DONE) {
    transmittedFlag = true;
  }
  else {
    receivedFlag = false;
    transmittedFlag = false;
  }
}

// ***********************************************************************************************************
void loop() 
{
  // Slave transmits  data to Master *************************************************************************
  Serial.println("Transmitting packet ... ");

  // battery voltage measurement
  int n = 256;     // average times
  int32_t Vadc = 0;
  for(int i = 0; i < n; i++) {
    Vadc = Vadc + analogRead(PIN_VBATT);
  }
  Vadc = Vadc / n;
  // sample map()data, need to calibrate each device
  int adc_3400 = 1442;  // ADC data when battery voltage is 3.4V
  int adc_4200 = 1633;  // ADC data when battery voltage is 4.2V  
  float Vbatt = map(Vadc, adc_3400, adc_4200, 3400, 4200) / 1000.0;   // Converts to battery voltage 3.4 ~ 4.2V

  // Slave transmits data
  ud.dataBuff32[0] = slaveuid;                    // Slave device UID
  ud.dataBuff32[1] = millis();                    // transmission time
  ud.dataBuff32F[2] = Vbatt;                      // battery voltage
  sec = ud.dataBuff32[1] / 1000;                  // transmission time sec

  digitalWrite(LED_RED, LOW);
  transmittedFlag = false;
  int state = radio.startTransmit(ud.dataBuff8, dataNum);
  digitalWrite(LED_RED, HIGH);

  txdata = "";
  for(int i = 0; i < dataNum; i++) {
    sprintf(printBuff, "%02x", ud.dataBuff8[i]);
    txdata += printBuff;
  }
  txdata.toUpperCase();
  Serial.println(txdata);

  Serial.print("Slave UID:\t"); Serial.println(slaveuid, HEX);
  Serial.print("Time:\t\t"); Serial.print(sec); Serial.println(" sec");
  Serial.print("Vbatt:\t\t"); Serial.print(Vbatt, 3); Serial.println(" V");

  display.clearDisplay();
  display.setCursor(0, 15); display.print("Slave");
  display.setCursor(64, 15); display.print(slaveuid, HEX);  
  display.setFont(&tahoma5pt7b);
  display.setCursor(0, 31); display.print(txdata);
  display.setFont(&tahoma7pt7b);
  display.setCursor(0, 63); display.print(sec); display.print(" sec");
  display.setCursor(64, 63); display.print(Vbatt, 3); display.print(" V");   
  display.display();

  // wait for transmittion completion
  while(!transmittedFlag) {
    while(SubGhz.isBusy());
  }

  // error status check
  if (state == RADIOLIB_ERR_NONE) {
    // packet was successfully sent
    Serial.println("transmission finished!");
  } 
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }

  // Slave receives response data from Master *********************************************************
  Serial.println("Waiting for incoming Masster response ... ");

  // start listening for LoRa packets
  state = radio.startReceive();
  if (state != RADIOLIB_ERR_NONE) {
    Serial.print("failed code ");
    Serial.println(state);
    while (true) {
      digitalWrite(LED_RED, LOW);
      delay(100);
      digitalWrite(LED_RED, HIGH);
      delay(100);
    }
  }

  // check if the flag is set
  while(!receivedFlag) {
    while(SubGhz.isBusy());
  }
  receivedFlag = false;

  // read a packet 
  state = radio.readData(ud.dataBuff8, dataNum);

  masteruid = ud.dataBuff32[0];   // Maste device UID
  if (state == RADIOLIB_ERR_NONE & masteruid == MASTER_UID) {
    // packet from Master was successfully received
    rxdata = "";
    for(int i = 0; i < dataNum; i++) {
      sprintf(printBuff, "%02x", ud.dataBuff8[i]);
      rxdata += printBuff;
    }
    rxdata.toUpperCase();
    Serial.println(rxdata);

    // restoration of received data
    rssi = ud.dataBuff16[2];  // RSSI
    snr = ud.dataBuff16[3];   // SNR

    Serial.print("Master UID:\t"); Serial.println(masteruid, HEX);
    Serial.print("Time:\t\t"); Serial.print(sec); Serial.println(" sec");
    Serial.print("RSSI:\t\t"); Serial.print(rssi); Serial.println(" dBm");
    Serial.print("SNR:\t\t"); Serial.print(snr); Serial.println(" dB");
    Serial.print("Vbatt:\t\t"); Serial.print(Vbatt, 3); Serial.println(" V");

    display.clearDisplay();
    display.setCursor(0, 15); display.print("from");
    display.setCursor(56, 15); display.print(masteruid, HEX);  
    display.setFont(&tahoma5pt7b);
    display.setCursor(0, 31); display.print(rxdata);
    display.setFont(&tahoma7pt7b);
    display.setCursor(10, 47); display.print(rssi); display.print(" dBm");
    display.setCursor(64, 47); display.print(snr); display.print(" dB");
    display.setCursor(0, 63); display.print(sec); display.print(" sec");
    display.setCursor(64, 63); display.print(Vbatt, 3); display.print(" V");   
    display.display();
  }
  // error status check
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }  

  // Sleep
  Serial.println("**** Sleep ****");
  Serial.println();
  Serial.flush();

  LowPower.shutdown(5000);
//  delay(5000);  
}
