// BSP : STM32 boards groups by STMicroelectronics 2.7.1
// Borad select : STM Boards groups / LoRa boars
// Board part number : LoRa-E5 mini (Board to be selected from Tools submenu 'Board part number)

// 1. wait for transmission of Slave
// 2. receives Slave data(Slave UID, mills(), Vbatt)
// 3. transtits response data(Master UID, rssi, snr, reserved)

// 2024/02/11

#include <RadioLib.h>           // 6.4.2  https://github.com/jgromes/RadioLib

// RM0461 Reference manual Rev 8  37.1.4 IEEE 64-bit unique device ID register (UID64)
#define MASTER_UID  *((uint32_t*)0x1FFF7580)    // use uid64 dor Master device UID

// \Arduino15\packages\STMicroelectronics\hardware\stm32\2.7.1
// \variants\STM32WLxx\WL54JCI_WL55JCI_WLE4J(8-B-C)I_WLE5J(8-B-C)I\variant_LORA_E5_MINI.h : 173
STM32WLx radio = new STM32WLx_Module();
static const uint32_t rfswitch_pins[] =
                         {PA4,  PA5,  RADIOLIB_NC};
static const Module::RfSwitchMode_t rfswitch_table[] = {
  {STM32WLx::MODE_IDLE,  {LOW,  LOW}},
  {STM32WLx::MODE_RX,    {HIGH, LOW}},
  {STM32WLx::MODE_TX_HP, {LOW, HIGH}},  // for LoRa-E5 mini
//  {STM32WLx::MODE_TX_LP, {HIGH, HIGH}},   // for LoRa-E5-LE mini
  END_OF_MODE_TABLE,
};

#define dataNum   12                // data bytes
union unionData {                   // for data type conversion
  uint32_t  dataBuff32[dataNum/4];
  float     dataBuff32F[dataNum/4];
  int16_t   dataBuff16[dataNum/2];
  uint8_t   dataBuff8[dataNum];
};
union unionData ud;

int rssi;               // signal RSSI [dBm]
int snr;                // signal SN ratio [dB]
uint32_t masteruid = MASTER_UID;    // Master device UID
uint32_t slaveuid;      // Slave device UID
uint32_t ms;            // timer [mS]
uint32_t us;            // timer [uS]
char printBuff[4];      // for sprintf()
String  txdata = "";    // transmission data packet string
String  rxdata = "";    // received data packet string
bool rxstatus = false;  // receive status flag

bool receivedFlag = false;     // flag that a packet was received
bool transmittedFlag = false;  // flag that a packet was transmitted

// *******************************************************************************************************
void setup() {
  Serial.begin(115200);
  while(!Serial);
  delay(2000);

  // Pin initialization
  pinMode(LED_RED, OUTPUT);         // built-in LED

  // STM32WL initialization
  radio.setRfSwitchTable(rfswitch_pins, rfswitch_table);

  // int16_t begin(float freq = 434.0, float bw = 125.0, uint8_t sf = 9, uint8_t cr = 7, \
  // uint8_t syncWord = RADIOLIB_SX126X_SYNC_WORD_PRIVATE, int8_t power = 10, \
  // uint16_t preambleLength = 8, float tcxoVoltage = 1.6, bool useRegulatorLDO = false);
  int state = radio.begin(922.6, 125.0, 7, 5, 0x12, 13, 8, 1.7, 0);

  if (state == RADIOLIB_ERR_NONE) {
    Serial.println("radio.begin() success!");
  }
  else {
    Serial.print("failed, code ");
    Serial.println(state);
    while(true) {
      digitalWrite(LED_RED, LOW);
      delay(100);
      digitalWrite(LED_RED, HIGH);
      delay(100);      
    }
  }

  // callback function when received or transmitted
  radio.setDio1Action(setFlag);
}

// callback function when received or transmitted
void setFlag(void) 
{
  uint16_t irqstatus = radio.getIrqStatus();
  if(irqstatus == RADIOLIB_SX126X_IRQ_RX_DONE) {
    receivedFlag = true;
  }
  else if(irqstatus == RADIOLIB_SX126X_IRQ_TX_DONE) {
    transmittedFlag = true;
  }
  else {
    receivedFlag = false;
    transmittedFlag = false;
  }
}

// ***********************************************************************************************************
void loop() 
{
  // Master receives packet from Slave *************************************************************************
  Serial.println("Waiting for incoming Slave transmitting packet ... ");

  // start listening for LoRa packets
  int state = radio.startReceive();
  if (state != RADIOLIB_ERR_NONE) {
    Serial.print("failed, code ");
    Serial.println(state);
    while (true) {
      digitalWrite(LED_RED, LOW);
      delay(100);
      digitalWrite(LED_RED, HIGH);
      delay(100);    
    }
  }

  // check if the flag is set
  while(!receivedFlag) {
    while(SubGhz.isBusy());
  }
  receivedFlag = false;

  // read a packet 
  state = radio.readData(ud.dataBuff8, dataNum);

  if (state == RADIOLIB_ERR_NONE) {   
    // packet form Slave was successfully received
    rxdata = "";
    for(int i = 0; i < dataNum; i++) {
      sprintf(printBuff, "%02x", ud.dataBuff8[i]);
      rxdata += printBuff;
    }
    rxdata.toUpperCase();
    Serial.println(rxdata);

    // restoration of received data
    slaveuid = ud.dataBuff32[0];    // Slave device UID    
    ms = ud.dataBuff32[1];          // timer [mS]
    us = ud.dataBuff32[2];          // timer [uS]    

    Serial.printf("Slave UID:\t%08x\n", slaveuid);
    Serial.print("Time:\t\t"); Serial.print(ms); Serial.println(" mS");
    Serial.print("Time:\t\t"); Serial.print(us); Serial.println(" uS");

    // RSSI and SNR of the last received packet
    Serial.print("RSSI:\t\t");
    rssi = radio.getRSSI();
    Serial.print(rssi);
    Serial.println(" dBm");

    Serial.print("SNR:\t\t");
    snr = radio.getSNR();
    Serial.print(snr);
    Serial.println(" dB");
  }
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }  

  delay(100); // need!

  // Master transmits response data to Slave  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Serial.println("Transmitting response packet ... ");

  // send data packet
  us = micros();
  ud.dataBuff32[0] = masteruid;     // Master device UID
  ud.dataBuff32[1] = us;            // timer [uS]
  ud.dataBuff16[4] = rssi;          // Master RSSI  dBm
  ud.dataBuff16[5] = snr;           // Master SN ratio  dB

  digitalWrite(LED_RED, LOW);
  transmittedFlag = false;
  state = radio.startTransmit(ud.dataBuff8, dataNum);
  digitalWrite(LED_RED, HIGH);

  // print the packet
  txdata = "";
  for(int i = 0; i < dataNum; i++) {
    sprintf(printBuff, "%02x", ud.dataBuff8[i]);
    txdata += printBuff;
  }
  txdata.toUpperCase();
  Serial.println(txdata);

  Serial.printf("Master UID:\t%08x\n", masteruid);
  Serial.print("Time:\t\t"); Serial.print(us); Serial.println(" uS");
  Serial.print("RSSI\t\t"); Serial.print(rssi); Serial.println(" dBm");
  Serial.print("SNR\t\t"); Serial.print(snr); Serial.println(" dB");
   
  // wait for transmittion completion
  while(!transmittedFlag) {
    while(SubGhz.isBusy());
  }

  // error status check
  if (state == RADIOLIB_ERR_NONE) {
    // packet was successfully sent
    Serial.println("transmission finished!");
    Serial.println();
  } 
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }
}

